"""
The module contains functions to launch SQL queries on Bauplan and retrieve
the result sets in a variety of formats (arrow Table, generator, file).
"""

import json
import os
from datetime import datetime
from typing import Any, Dict, Generator, Optional

import grpc
import pyarrow as pa
import pyarrow.flight as flight

from . import exceptions
from ._common import BAUPLAN_VERSION, _get_commander_and_metadata, _get_or_validate_branch
from ._protobufs.bauplan_pb2 import TriggerRunRequest


def _iterate_flight_stream_batches(
    reader: flight.FlightStreamReader, max_rows: Optional[int] = None
) -> Generator[pa.lib.RecordBatch, None, None]:
    rows = 0
    try:
        if reader is None:
            raise exceptions.NoResultsFoundError('No results found')
        chunk: Optional[pa.lib.RecordBatch] = reader.read_chunk()
        if chunk is not None:
            batch: pa.lib.RecordBatch = chunk.data
            yield batch
            if max_rows:
                rows += batch.num_rows
                if rows >= max_rows:
                    raise StopIteration
        else:
            raise StopIteration
    except StopIteration:
        pass


def _query(
    api_key: str | None,
    query: str,
    max_rows: Optional[int] = None,
    no_cache: bool = False,
    branch_name: str = 'main',
    connector: Optional[str] = None,
    connector_config_key: Optional[str] = None,
    connector_config_uri: Optional[str] = None,
    return_flight_stream: Optional[bool] = None,
    args: Optional[Dict[str, Any]] = None,
) -> pa.Table:
    """
    Execute a SQL query and return the results as a pyarrow.Table.

    If you prefer to return the raw FlightStreamReader, pass `return_flight_stream=True`.
    """
    if max_rows is not None:
        # max_rows limits
        if not isinstance(max_rows, int) or not (0 < max_rows < 100000000):
            raise ValueError('max_rows must be positive integer 1-100000000')
    branch_name = _get_or_validate_branch(branch_name)

    if args is not None and not isinstance(args, dict):
        raise ValueError('args must be a dict or None')
    if args:
        if branch_name is not None and 'read-branch' in args:
            raise ValueError('can only pass one of branch_name and args["read-branch"]')
        if 'read-branch' in args:
            branch_name = _get_or_validate_branch(args['read-branch'])

    # rebuild a query with the connector strings if specified
    # note that if the connector is not specified we get back the query as is
    query = _add_connector_strings_to_query(query, connector, connector_config_key, connector_config_uri)
    trigger_run_request: TriggerRunRequest = TriggerRunRequest(
        module_version=BAUPLAN_VERSION,
        args=args or {},
        query_for_flight=query,
        is_flight_query=True,
    )
    # else:
    trigger_run_request.args['read-branch'] = branch_name
    if no_cache:
        trigger_run_request.args['runner-cache'] = 'off'
    client, metadata = _get_commander_and_metadata(api_key)
    job_id: TriggerRunRequest = client.TriggerRun(trigger_run_request, metadata=metadata)
    log_stream: grpc.Call = client.SubscribeLogs(job_id, metadata=metadata)
    flight_endpoint: Optional[str] = None
    for log in log_stream:
        if os.getenv('BPLN_DEBUG'):
            print(log)

        ev = log.runner_event
        if ev and ev.WhichOneof('event') == 'flight_server_start':
            flight_endpoint = log.runner_event.flight_server_start.endpoint
            break
    if not flight_endpoint:
        return None
    flight_client: flight.FlightClient = flight.FlightClient('grpc://' + flight_endpoint)
    options: flight.FlightCallOptions = flight.FlightCallOptions(
        headers=[(b'authorization', 'Bearer my_special_token'.encode())]
    )
    ticket: flight.Ticket = next(flight_client.list_flights(options=options)).endpoints[0].ticket
    reader: flight.FlightStreamReader = flight_client.do_get(ticket, options=options)

    if return_flight_stream:
        return reader
    if reader is None:
        raise exceptions.NoResultsFoundError('No results found')
    if max_rows:
        num_rows = 0
        data = None
        try:
            for batch in _iterate_flight_stream_batches(reader, max_rows):
                for i in range(batch.num_rows):
                    row = _row_to_dict(batch, i, batch.schema)
                    if not data:
                        data = {k: [v] for k, v in row.items()}
                    else:
                        for k, v in row.items():
                            data[k].append(v)
                    if max_rows:
                        num_rows += 1
                        if num_rows >= max_rows:
                            raise StopIteration
        except StopIteration:
            pass
        return pa.table({k: v[:max_rows] for k, v in data.items()}, schema=batch.schema)

    return reader.read_all()


def _add_connector_strings_to_query(
    query: str,
    connector: Optional[str] = None,
    connector_config_key: Optional[str] = None,
    connector_config_uri: Optional[str] = None,
) -> str:
    """

    Add the connector strings to the query to allow the backend to direct the query to the correct engine.
    We assume that if the connector is not specified we use Bauplan as is; the other properties default to
    sensible values (check the docs for the details!).

    """
    if not connector:
        return query

    connector_string = f'-- bauplan: connector={connector}'
    connector_config_key_string = (
        f'-- bauplan: connector.config_key={connector_config_key}' if connector_config_key else ''
    )
    connector_config_uri_string = (
        f'-- bauplan: connector.config_uri={connector_config_uri}' if connector_config_uri else ''
    )

    return f'{connector_string}\n{connector_config_key_string}\n{connector_config_uri_string}\n{query}'


def _build_query_from_scan(
    table_name: str,
    columns: Optional[list] = None,
    filters: Optional[str] = None,
    limit: Optional[int] = None,
) -> str:
    """
    Take as input the arguments of the scan function and build a SQL query
    using SQLGlot.

    :meta private:

    """
    from sqlglot import select

    cols = columns or ['*']
    q = select(*cols).from_(table_name).where(filters)
    if limit:
        q = q.limit(limit)

    return q.sql()


def _scan(
    api_key: str | None,
    table_name: str,
    columns: Optional[list] = None,
    filters: Optional[str] = None,
    limit: Optional[int] = None,
    branch_name: str = 'main',
    connector: Optional[str] = None,
    connector_config_key: Optional[str] = None,
    connector_config_uri: Optional[str] = None,
    args: Optional[Dict] | None = None,
    **kwargs: Any,
) -> pa.Table:
    """
    Execute a table scan (with optional filters) and return the results as an arrow Table.
    Note that this function uses SQLGlot to compose a safe SQL query,
    and then internally defer to the query_to_arrow function for the actual scan.
    """
    q = _build_query_from_scan(table_name, columns, filters, limit)
    return _query(
        api_key=api_key,
        query=q,
        branch_name=branch_name,
        connector=connector,
        connector_config_key=connector_config_key,
        connector_config_uri=connector_config_uri,
        args=args,
        **kwargs,
    )


def _query_to_generator(
    api_key: str | None,
    query: str,
    branch_name: Optional[str] | None = None,
    no_cache: Optional[bool] = False,
    max_rows: Optional[int] = None,
    connector: Optional[str] = None,
    connector_config_key: Optional[str] = None,
    connector_config_uri: Optional[str] = None,
    args: Optional[Dict[str, Any]] = None,
) -> Generator[Dict[str, Any], None, None]:
    """
    Execute a SQL query and return the results as a generator, where each row is
    a Python dictionary.
    """
    as_json = False
    if args and args.get('as_json'):
        as_json = True
        del args['as_json']
    reader: flight.FlightStreamReader = _query(
        api_key=api_key,
        query=query,
        max_rows=max_rows,
        no_cache=no_cache,
        branch_name=branch_name,
        connector=connector,
        connector_config_key=connector_config_key,
        connector_config_uri=connector_config_uri,
        return_flight_stream=True,
        args=args,
    )
    if reader is None:
        raise exceptions.NoResultsFoundError('No results found')
    as_json = False
    if args and args.get('as_json'):
        as_json = True
    num_rows = 0
    try:
        for batch in _iterate_flight_stream_batches(reader, max_rows):
            for i in range(batch.num_rows):
                yield _row_to_dict(batch, i, batch.schema, as_json=as_json)
                if max_rows:
                    num_rows += 1
                    if num_rows >= max_rows:
                        raise StopIteration
    except StopIteration:
        pass


def _row_to_dict(
    batch: pa.lib.RecordBatch, row_index: int, schema: pa.lib.Schema, as_json: Optional[bool] = False
) -> Dict[str, Any]:
    """
    Convert a row of a ``pyarrow.RecordBatch`` to a dictionary.

    :meta private:

    :param batch: The ``pyarrow.RecordBatch`` containing the row.
    :param row_index: The index of the row to convert.
    :param schema: The schema of the ``RecordBatch``.
    :param as_json: Whether or not to cast to JSON-compatible types (i.e. datetime -> ISO format string).
    :return: A dictionary representing the row.
    """
    row: Dict[str, Any] = {}
    for j, name in enumerate(schema.names):
        column: pa.lib.ChunkedArray = batch.column(j)
        value = column[row_index].as_py()
        if as_json:
            if isinstance(value, datetime):
                value = value.isoformat()
        row[name] = value
    return row


def _query_to_file(
    api_key: str | None,
    filename: str,
    query: str,
    max_rows: Optional[int] = None,
    no_cache: Optional[bool] = False,
    branch_name: str = 'main',
    connector: Optional[str] = None,
    connector_config_key: Optional[str] = None,
    connector_config_uri: Optional[str] = None,
    args: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Execute a SQL query and write the results to a file.
    """
    if filename.endswith('.json'):
        if not args:
            args = {}
        args['as_json'] = True
        with open(filename, 'w') as outfile:
            outfile.write('[\n')
            first_row: bool = True
            for row in _query_to_generator(
                api_key=api_key,
                query=query,
                max_rows=max_rows,
                no_cache=no_cache,
                branch_name=branch_name,
                connector=connector,
                connector_config_key=connector_config_key,
                connector_config_uri=connector_config_uri,
                args=args,
            ):
                if not first_row:
                    outfile.write(',\n')
                    first_row = False
                outfile.write(json.dumps(row))
            outfile.write('\n]')
    else:
        raise ValueError('Only .json extension is supported for filename')
